#![feature(proc_macro_hygiene, decl_macro)]

#[macro_use] extern crate rocket;
#[macro_use] extern crate rocket_contrib;
#[macro_use] extern crate serde_derive;

use rocket::State;
use rocket_contrib::json::Json;
use pathfinding::grid::Grid;
use std::collections::HashMap;
use std::sync::{Arc, Mutex};

mod battlesnake;

type Cache = Arc<Mutex<HashMap<String, GameState>>>;
pub struct GameState {
    path: Option<Vec<battlesnake::Coord>>,
    target: battlesnake::Coord,
}

#[get("/")]
fn index() -> &'static str {
    "Hello, world!"
}

#[get("/ping")]
fn ping() -> &'static str {
    "OK - Pneuma online"
}

#[post("/begin", format = "json", data = "<msg>")]
fn begin(cache: State<Cache>, msg: Json<battlesnake::SnakeRequest>) -> Json<battlesnake::StartResponse> {
    let head = msg.you.body[0];
    let target = find_target(&msg);
    let path = find_path(&msg, &head, &target);
    let gs = GameState{
        target: *target,
        path: path,
    };

    cache.inner().get_mut().unwrap().insert(msg.game.id, gs);

    Json(battlesnake::StartResponse{
        color: "#5ce8c3".to_string(),
        head_type: "beluga".to_string(),
        tail_type: "skinny".to_string(),
    })
}

fn find_path(msg: &battlesnake::SnakeRequest, head: &battlesnake::Coord, target: &battlesnake::Coord) -> Option<Vec<battlesnake::Coord>> {
    let path = pathfinding::directed::astar::astar(
        head,
        |n| msg.board.safe_neighbors(n).into_iter(),
        |n| (battlesnake::manhattan(n, &target) as usize),
        |n| n == target,
    );

    match path {
        None => return None,
        Some(x) => {
            return Some(x.0);
        },
    }
}

#[post("/move", format = "json", data = "<msg>")]
fn make_move(msg: Json<battlesnake::SnakeRequest>) -> Json<battlesnake::MoveResponse> {
    let target = find_target(&msg);
    println!("target: {:?}", target);

    let head = &msg.you.body[0];
    println!("me:     {:?}", head);
    for ne in &msg.board.safe_neighbors(&head) {
        println!("{:?} {}", ne.0, battlesnake::Line{
            start: &head,
            end: &ne.0,
        }.direction());
    }

    println!("{:?}", pathfinding::directed::astar::astar(
        head,
        |n| msg.board.safe_neighbors(n).into_iter(),
        |n| (battlesnake::manhattan(n, &target) as usize),
        |n| n == target,
    ));

    Json(battlesnake::MoveResponse {
        move_field: "up".to_string(),
    })
}

fn find_target<'a>(gs: &'a battlesnake::SnakeRequest) -> &'a battlesnake::Coord {
    let head = &gs.you.body[0];
    if gs.you.health > 30 {
        let mut lowest_score: u32 = 99999;
        let mut coord: &battlesnake::Coord = &gs.you.body.last().unwrap();

        for food in &gs.board.food {
            let score = battlesnake::manhattan(&head, &food);
            if score < lowest_score {
                lowest_score = score;
                coord = food;
            }
        }

        return coord;
    }

    return gs.you.body.last().unwrap();
}

fn main() {
    let map = HashMap::<String, GameState>::new();
    let mutex_map = Mutex::from(map);
    let arc_mutex_map = Arc::from(mutex_map);
    rocket::ignite().mount("/", routes![
        index,
        begin,
        ping,
        make_move,
    ])
        .manage(arc_mutex_map)
        .launch();
}
